# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: relationalai/lqp/v1/logic.proto
# Protobuf Python Version: 4.25.6
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1frelationalai/lqp/v1/logic.proto\x12\x13relationalai.lqp.v1\"\x8f\x01\n\x0b\x44\x65\x63laration\x12,\n\x03\x64\x65\x66\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.DefH\x00R\x03\x64\x65\x66\x12>\n\talgorithm\x18\x02 \x01(\x0b\x32\x1e.relationalai.lqp.v1.AlgorithmH\x00R\talgorithmB\x12\n\x10\x64\x65\x63laration_type\"\xa6\x01\n\x03\x44\x65\x66\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x03 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"u\n\tAlgorithm\x12\x37\n\x06global\x18\x01 \x03(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x06global\x12/\n\x04\x62ody\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ScriptR\x04\x62ody\"H\n\x06Script\x12>\n\nconstructs\x18\x01 \x03(\x0b\x32\x1e.relationalai.lqp.v1.ConstructR\nconstructs\"\x94\x01\n\tConstruct\x12/\n\x04loop\x18\x01 \x01(\x0b\x32\x19.relationalai.lqp.v1.LoopH\x00R\x04loop\x12\x44\n\x0binstruction\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.InstructionH\x00R\x0binstructionB\x10\n\x0e\x63onstruct_type\"m\n\x04Loop\x12\x34\n\x04init\x18\x01 \x03(\x0b\x32 .relationalai.lqp.v1.InstructionR\x04init\x12/\n\x04\x62ody\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ScriptR\x04\x62ody\"\xed\x02\n\x0bInstruction\x12\x35\n\x06\x61ssign\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.AssignH\x00R\x06\x61ssign\x12\x35\n\x06upsert\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.UpsertH\x00R\x06upsert\x12\x32\n\x05\x62reak\x18\x03 \x01(\x0b\x32\x1a.relationalai.lqp.v1.BreakH\x00R\x05\x62reak\x12/\n\x04\x63opy\x18\x04 \x01(\x0b\x32\x19.relationalai.lqp.v1.CopyH\x00R\x04\x63opy\x12?\n\nmonoid_def\x18\x05 \x01(\x0b\x32\x1e.relationalai.lqp.v1.MonoidDefH\x00R\tmonoidDef\x12<\n\tmonus_def\x18\x06 \x01(\x0b\x32\x1d.relationalai.lqp.v1.MonusDefH\x00R\x08monusDefB\x0c\n\ninstr_type\"\xa9\x01\n\x06\x41ssign\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x03 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\xa9\x01\n\x06Upsert\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x03 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\xa8\x01\n\x05\x42reak\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x03 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\xa7\x01\n\x04\x43opy\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x03 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\xe1\x01\n\tMonoidDef\x12\x33\n\x06monoid\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.MonoidR\x06monoid\x12\x33\n\x04name\x18\x02 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x04 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\xe0\x01\n\x08MonusDef\x12\x33\n\x06monoid\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.MonoidR\x06monoid\x12\x33\n\x04name\x18\x02 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x04 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\x92\x02\n\x06Monoid\x12<\n\tor_monoid\x18\x01 \x01(\x0b\x32\x1d.relationalai.lqp.v1.OrMonoidH\x00R\x08orMonoid\x12?\n\nmin_monoid\x18\x02 \x01(\x0b\x32\x1e.relationalai.lqp.v1.MinMonoidH\x00R\tminMonoid\x12?\n\nmax_monoid\x18\x03 \x01(\x0b\x32\x1e.relationalai.lqp.v1.MaxMonoidH\x00R\tmaxMonoid\x12?\n\nsum_monoid\x18\x04 \x01(\x0b\x32\x1e.relationalai.lqp.v1.SumMonoidH\x00R\tsumMonoidB\x07\n\x05value\"\n\n\x08OrMonoid\":\n\tMinMonoid\x12-\n\x04type\x18\x01 \x01(\x0b\x32\x19.relationalai.lqp.v1.TypeR\x04type\":\n\tMaxMonoid\x12-\n\x04type\x18\x01 \x01(\x0b\x32\x19.relationalai.lqp.v1.TypeR\x04type\":\n\tSumMonoid\x12-\n\x04type\x18\x01 \x01(\x0b\x32\x19.relationalai.lqp.v1.TypeR\x04type\"d\n\x07\x42inding\x12*\n\x03var\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.VarR\x03var\x12-\n\x04type\x18\x02 \x01(\x0b\x32\x19.relationalai.lqp.v1.TypeR\x04type\"s\n\x0b\x41\x62straction\x12\x30\n\x04vars\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.BindingR\x04vars\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x05value\"\x83\x05\n\x07\x46ormula\x12\x35\n\x06\x65xists\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ExistsH\x00R\x06\x65xists\x12\x35\n\x06reduce\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ReduceH\x00R\x06reduce\x12\x44\n\x0b\x63onjunction\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.ConjunctionH\x00R\x0b\x63onjunction\x12\x44\n\x0b\x64isjunction\x18\x04 \x01(\x0b\x32 .relationalai.lqp.v1.DisjunctionH\x00R\x0b\x64isjunction\x12,\n\x03not\x18\x05 \x01(\x0b\x32\x18.relationalai.lqp.v1.NotH\x00R\x03not\x12,\n\x03\x66\x66i\x18\x06 \x01(\x0b\x32\x18.relationalai.lqp.v1.FFIH\x00R\x03\x66\x66i\x12/\n\x04\x61tom\x18\x07 \x01(\x0b\x32\x19.relationalai.lqp.v1.AtomH\x00R\x04\x61tom\x12\x35\n\x06pragma\x18\x08 \x01(\x0b\x32\x1b.relationalai.lqp.v1.PragmaH\x00R\x06pragma\x12>\n\tprimitive\x18\t \x01(\x0b\x32\x1e.relationalai.lqp.v1.PrimitiveH\x00R\tprimitive\x12\x39\n\x08rel_atom\x18\n \x01(\x0b\x32\x1c.relationalai.lqp.v1.RelAtomH\x00R\x07relAtom\x12/\n\x04\x63\x61st\x18\x0b \x01(\x0b\x32\x19.relationalai.lqp.v1.CastH\x00R\x04\x63\x61stB\x0e\n\x0c\x66ormula_type\">\n\x06\x45xists\x12\x34\n\x04\x62ody\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\"\xa1\x01\n\x06Reduce\x12\x30\n\x02op\x18\x01 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x02op\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12/\n\x05terms\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"?\n\x0b\x43onjunction\x12\x30\n\x04\x61rgs\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x04\x61rgs\"?\n\x0b\x44isjunction\x12\x30\n\x04\x61rgs\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x04\x61rgs\"5\n\x03Not\x12.\n\x03\x61rg\x18\x01 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x03\x61rg\"\x80\x01\n\x03\x46\x46I\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x34\n\x04\x61rgs\x18\x02 \x03(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x61rgs\x12/\n\x05terms\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"l\n\x04\x41tom\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"M\n\x06Pragma\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"S\n\tPrimitive\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x32\n\x05terms\x18\x02 \x03(\x0b\x32\x1c.relationalai.lqp.v1.RelTermR\x05terms\"Q\n\x07RelAtom\x12\x12\n\x04name\x18\x03 \x01(\tR\x04name\x12\x32\n\x05terms\x18\x02 \x03(\x0b\x32\x1c.relationalai.lqp.v1.RelTermR\x05terms\"j\n\x04\x43\x61st\x12/\n\x05input\x18\x02 \x01(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05input\x12\x31\n\x06result\x18\x03 \x01(\x0b\x32\x19.relationalai.lqp.v1.TermR\x06result\"\x96\x01\n\x07RelTerm\x12I\n\x11specialized_value\x18\x01 \x01(\x0b\x32\x1a.relationalai.lqp.v1.ValueH\x00R\x10specializedValue\x12/\n\x04term\x18\x02 \x01(\x0b\x32\x19.relationalai.lqp.v1.TermH\x00R\x04termB\x0f\n\rrel_term_type\"{\n\x04Term\x12,\n\x03var\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.VarH\x00R\x03var\x12\x38\n\x08\x63onstant\x18\x02 \x01(\x0b\x32\x1a.relationalai.lqp.v1.ValueH\x00R\x08\x63onstantB\x0b\n\tterm_type\"\x19\n\x03Var\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\"O\n\tAttribute\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12.\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x1a.relationalai.lqp.v1.ValueR\x04\x61rgs\"<\n\nRelationId\x12\x15\n\x06id_low\x18\x01 \x01(\x06R\x05idLow\x12\x17\n\x07id_high\x18\x02 \x01(\x06R\x06idHigh\"\x89\x06\n\x04Type\x12Q\n\x10unspecified_type\x18\x01 \x01(\x0b\x32$.relationalai.lqp.v1.UnspecifiedTypeH\x00R\x0funspecifiedType\x12\x42\n\x0bstring_type\x18\x02 \x01(\x0b\x32\x1f.relationalai.lqp.v1.StringTypeH\x00R\nstringType\x12\x39\n\x08int_type\x18\x03 \x01(\x0b\x32\x1c.relationalai.lqp.v1.IntTypeH\x00R\x07intType\x12?\n\nfloat_type\x18\x04 \x01(\x0b\x32\x1e.relationalai.lqp.v1.FloatTypeH\x00R\tfloatType\x12\x45\n\x0cuint128_type\x18\x05 \x01(\x0b\x32 .relationalai.lqp.v1.UInt128TypeH\x00R\x0buint128Type\x12\x42\n\x0bint128_type\x18\x06 \x01(\x0b\x32\x1f.relationalai.lqp.v1.Int128TypeH\x00R\nint128Type\x12<\n\tdate_type\x18\x07 \x01(\x0b\x32\x1d.relationalai.lqp.v1.DateTypeH\x00R\x08\x64\x61teType\x12H\n\rdatetime_type\x18\x08 \x01(\x0b\x32!.relationalai.lqp.v1.DateTimeTypeH\x00R\x0c\x64\x61tetimeType\x12\x45\n\x0cmissing_type\x18\t \x01(\x0b\x32 .relationalai.lqp.v1.MissingTypeH\x00R\x0bmissingType\x12\x45\n\x0c\x64\x65\x63imal_type\x18\n \x01(\x0b\x32 .relationalai.lqp.v1.DecimalTypeH\x00R\x0b\x64\x65\x63imalType\x12\x45\n\x0c\x62oolean_type\x18\x0b \x01(\x0b\x32 .relationalai.lqp.v1.BooleanTypeH\x00R\x0b\x62ooleanTypeB\x06\n\x04type\"\x11\n\x0fUnspecifiedType\"\x0c\n\nStringType\"\t\n\x07IntType\"\x0b\n\tFloatType\"\r\n\x0bUInt128Type\"\x0c\n\nInt128Type\"\n\n\x08\x44\x61teType\"\x0e\n\x0c\x44\x61teTimeType\"\r\n\x0bMissingType\"A\n\x0b\x44\x65\x63imalType\x12\x1c\n\tprecision\x18\x01 \x01(\x05R\tprecision\x12\x14\n\x05scale\x18\x02 \x01(\x05R\x05scale\"\r\n\x0b\x42ooleanType\"\xd1\x04\n\x05Value\x12#\n\x0cstring_value\x18\x01 \x01(\tH\x00R\x0bstringValue\x12\x1d\n\tint_value\x18\x02 \x01(\x03H\x00R\x08intValue\x12!\n\x0b\x66loat_value\x18\x03 \x01(\x01H\x00R\nfloatValue\x12H\n\ruint128_value\x18\x04 \x01(\x0b\x32!.relationalai.lqp.v1.UInt128ValueH\x00R\x0cuint128Value\x12\x45\n\x0cint128_value\x18\x05 \x01(\x0b\x32 .relationalai.lqp.v1.Int128ValueH\x00R\x0bint128Value\x12H\n\rmissing_value\x18\x06 \x01(\x0b\x32!.relationalai.lqp.v1.MissingValueH\x00R\x0cmissingValue\x12?\n\ndate_value\x18\x07 \x01(\x0b\x32\x1e.relationalai.lqp.v1.DateValueH\x00R\tdateValue\x12K\n\x0e\x64\x61tetime_value\x18\x08 \x01(\x0b\x32\".relationalai.lqp.v1.DateTimeValueH\x00R\rdatetimeValue\x12H\n\rdecimal_value\x18\t \x01(\x0b\x32!.relationalai.lqp.v1.DecimalValueH\x00R\x0c\x64\x65\x63imalValue\x12%\n\rboolean_value\x18\n \x01(\x08H\x00R\x0c\x62ooleanValueB\x07\n\x05value\"4\n\x0cUInt128Value\x12\x10\n\x03low\x18\x01 \x01(\x06R\x03low\x12\x12\n\x04high\x18\x02 \x01(\x06R\x04high\"3\n\x0bInt128Value\x12\x10\n\x03low\x18\x01 \x01(\x06R\x03low\x12\x12\n\x04high\x18\x02 \x01(\x06R\x04high\"\x0e\n\x0cMissingValue\"G\n\tDateValue\x12\x12\n\x04year\x18\x01 \x01(\x05R\x04year\x12\x14\n\x05month\x18\x02 \x01(\x05R\x05month\x12\x10\n\x03\x64\x61y\x18\x03 \x01(\x05R\x03\x64\x61y\"\xb1\x01\n\rDateTimeValue\x12\x12\n\x04year\x18\x01 \x01(\x05R\x04year\x12\x14\n\x05month\x18\x02 \x01(\x05R\x05month\x12\x10\n\x03\x64\x61y\x18\x03 \x01(\x05R\x03\x64\x61y\x12\x12\n\x04hour\x18\x04 \x01(\x05R\x04hour\x12\x16\n\x06minute\x18\x05 \x01(\x05R\x06minute\x12\x16\n\x06second\x18\x06 \x01(\x05R\x06second\x12 \n\x0bmicrosecond\x18\x07 \x01(\x05R\x0bmicrosecond\"z\n\x0c\x44\x65\x63imalValue\x12\x1c\n\tprecision\x18\x01 \x01(\x05R\tprecision\x12\x14\n\x05scale\x18\x02 \x01(\x05R\x05scale\x12\x36\n\x05value\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.Int128ValueR\x05valueb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'relationalai.lqp.v1.logic_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_DECLARATION']._serialized_start=57
  _globals['_DECLARATION']._serialized_end=200
  _globals['_DEF']._serialized_start=203
  _globals['_DEF']._serialized_end=369
  _globals['_ALGORITHM']._serialized_start=371
  _globals['_ALGORITHM']._serialized_end=488
  _globals['_SCRIPT']._serialized_start=490
  _globals['_SCRIPT']._serialized_end=562
  _globals['_CONSTRUCT']._serialized_start=565
  _globals['_CONSTRUCT']._serialized_end=713
  _globals['_LOOP']._serialized_start=715
  _globals['_LOOP']._serialized_end=824
  _globals['_INSTRUCTION']._serialized_start=827
  _globals['_INSTRUCTION']._serialized_end=1192
  _globals['_ASSIGN']._serialized_start=1195
  _globals['_ASSIGN']._serialized_end=1364
  _globals['_UPSERT']._serialized_start=1367
  _globals['_UPSERT']._serialized_end=1536
  _globals['_BREAK']._serialized_start=1539
  _globals['_BREAK']._serialized_end=1707
  _globals['_COPY']._serialized_start=1710
  _globals['_COPY']._serialized_end=1877
  _globals['_MONOIDDEF']._serialized_start=1880
  _globals['_MONOIDDEF']._serialized_end=2105
  _globals['_MONUSDEF']._serialized_start=2108
  _globals['_MONUSDEF']._serialized_end=2332
  _globals['_MONOID']._serialized_start=2335
  _globals['_MONOID']._serialized_end=2609
  _globals['_ORMONOID']._serialized_start=2611
  _globals['_ORMONOID']._serialized_end=2621
  _globals['_MINMONOID']._serialized_start=2623
  _globals['_MINMONOID']._serialized_end=2681
  _globals['_MAXMONOID']._serialized_start=2683
  _globals['_MAXMONOID']._serialized_end=2741
  _globals['_SUMMONOID']._serialized_start=2743
  _globals['_SUMMONOID']._serialized_end=2801
  _globals['_BINDING']._serialized_start=2803
  _globals['_BINDING']._serialized_end=2903
  _globals['_ABSTRACTION']._serialized_start=2905
  _globals['_ABSTRACTION']._serialized_end=3020
  _globals['_FORMULA']._serialized_start=3023
  _globals['_FORMULA']._serialized_end=3666
  _globals['_EXISTS']._serialized_start=3668
  _globals['_EXISTS']._serialized_end=3730
  _globals['_REDUCE']._serialized_start=3733
  _globals['_REDUCE']._serialized_end=3894
  _globals['_CONJUNCTION']._serialized_start=3896
  _globals['_CONJUNCTION']._serialized_end=3959
  _globals['_DISJUNCTION']._serialized_start=3961
  _globals['_DISJUNCTION']._serialized_end=4024
  _globals['_NOT']._serialized_start=4026
  _globals['_NOT']._serialized_end=4079
  _globals['_FFI']._serialized_start=4082
  _globals['_FFI']._serialized_end=4210
  _globals['_ATOM']._serialized_start=4212
  _globals['_ATOM']._serialized_end=4320
  _globals['_PRAGMA']._serialized_start=4322
  _globals['_PRAGMA']._serialized_end=4399
  _globals['_PRIMITIVE']._serialized_start=4401
  _globals['_PRIMITIVE']._serialized_end=4484
  _globals['_RELATOM']._serialized_start=4486
  _globals['_RELATOM']._serialized_end=4567
  _globals['_CAST']._serialized_start=4569
  _globals['_CAST']._serialized_end=4675
  _globals['_RELTERM']._serialized_start=4678
  _globals['_RELTERM']._serialized_end=4828
  _globals['_TERM']._serialized_start=4830
  _globals['_TERM']._serialized_end=4953
  _globals['_VAR']._serialized_start=4955
  _globals['_VAR']._serialized_end=4980
  _globals['_ATTRIBUTE']._serialized_start=4982
  _globals['_ATTRIBUTE']._serialized_end=5061
  _globals['_RELATIONID']._serialized_start=5063
  _globals['_RELATIONID']._serialized_end=5123
  _globals['_TYPE']._serialized_start=5126
  _globals['_TYPE']._serialized_end=5903
  _globals['_UNSPECIFIEDTYPE']._serialized_start=5905
  _globals['_UNSPECIFIEDTYPE']._serialized_end=5922
  _globals['_STRINGTYPE']._serialized_start=5924
  _globals['_STRINGTYPE']._serialized_end=5936
  _globals['_INTTYPE']._serialized_start=5938
  _globals['_INTTYPE']._serialized_end=5947
  _globals['_FLOATTYPE']._serialized_start=5949
  _globals['_FLOATTYPE']._serialized_end=5960
  _globals['_UINT128TYPE']._serialized_start=5962
  _globals['_UINT128TYPE']._serialized_end=5975
  _globals['_INT128TYPE']._serialized_start=5977
  _globals['_INT128TYPE']._serialized_end=5989
  _globals['_DATETYPE']._serialized_start=5991
  _globals['_DATETYPE']._serialized_end=6001
  _globals['_DATETIMETYPE']._serialized_start=6003
  _globals['_DATETIMETYPE']._serialized_end=6017
  _globals['_MISSINGTYPE']._serialized_start=6019
  _globals['_MISSINGTYPE']._serialized_end=6032
  _globals['_DECIMALTYPE']._serialized_start=6034
  _globals['_DECIMALTYPE']._serialized_end=6099
  _globals['_BOOLEANTYPE']._serialized_start=6101
  _globals['_BOOLEANTYPE']._serialized_end=6114
  _globals['_VALUE']._serialized_start=6117
  _globals['_VALUE']._serialized_end=6710
  _globals['_UINT128VALUE']._serialized_start=6712
  _globals['_UINT128VALUE']._serialized_end=6764
  _globals['_INT128VALUE']._serialized_start=6766
  _globals['_INT128VALUE']._serialized_end=6817
  _globals['_MISSINGVALUE']._serialized_start=6819
  _globals['_MISSINGVALUE']._serialized_end=6833
  _globals['_DATEVALUE']._serialized_start=6835
  _globals['_DATEVALUE']._serialized_end=6906
  _globals['_DATETIMEVALUE']._serialized_start=6909
  _globals['_DATETIMEVALUE']._serialized_end=7086
  _globals['_DECIMALVALUE']._serialized_start=7088
  _globals['_DECIMALVALUE']._serialized_end=7210
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Union, Tuple, Sequence, Optional, Dict
import datetime as dt
from decimal import Decimal

# Tree representation of LQP. Each non-terminal (those with more than one
# option) is an "abstract" class and each terminal is its own class. All of
# which are children of LqpNode. Value is an exception -- it is just a value.

@dataclass(frozen=True)
class SourceInfo:
    file: str
    line: int
    column: int

    def __str__(self) -> str:
        return f"{self.file}:{self.line}:{self.column}"

# --- Logic Types ---

@dataclass(frozen=True)
class LqpNode:
    meta: Optional[SourceInfo]

# Declaration := Def | Algorithm
@dataclass(frozen=True)
class Declaration(LqpNode):
    pass

# Def(name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class Def(Declaration):
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

# Algorithm(globals::RelationId[], body::Script)
@dataclass(frozen=True)
class Algorithm(Declaration):
    global_: Sequence[RelationId]
    body: Script

# Script := Construct[]
@dataclass(frozen=True)
class Script(LqpNode):
    constructs: Sequence[Construct]

# Construct := Loop | Instruction
@dataclass(frozen=True)
class Construct(LqpNode):
    pass

# Loop(init::Instruction[], body::Algorithm)
@dataclass(frozen=True)
class Loop(Construct):
    init: Sequence[Instruction]
    body: Script

# Instruction := Assign | Break | Upsert | Copy | MonoidDef | MonusDef
@dataclass(frozen=True)
class Instruction(Construct):
    pass

# Assign(name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class Assign(Instruction):
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

# Upsert(name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class Upsert(Instruction):
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

# Break(name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class Break(Instruction):
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

# Copy(name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class Copy(Instruction):
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

# MonoidDef(monoid::Monoid, name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class MonoidDef(Instruction):
    monoid: Monoid
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

# MonusDef(monoid::Monoid, name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class MonusDef(Instruction):
    monoid: Monoid
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

# Monoid := OrMonoid | MinMonoid | MaxMonoid | SumMonoid
@dataclass(frozen=True)
class Monoid(LqpNode):
    pass

# OrMonoid
@dataclass(frozen=True)
class OrMonoid(Monoid):
    pass

# MinMonoid
@dataclass(frozen=True)
class MinMonoid(Monoid):
    type: Type

# MaxMonoid
@dataclass(frozen=True)
class MaxMonoid(Monoid):
    type: Type

# SumMonoid
@dataclass(frozen=True)
class SumMonoid(Monoid):
    type: Type

# Abstraction(vars::Binding[], value::Formula)
@dataclass(frozen=True)
class Abstraction(LqpNode):
    vars: Sequence[Tuple[Var, Type]]
    value: Formula

# Formula := Exists | Reduce | Conjunction | Disjunction | Not | FFI | Atom | Pragma | Primitive | TrueVal | FalseVal | RelAtom | Cast
@dataclass(frozen=True)
class Formula(LqpNode):
    pass

# Exists(body::Abstraction)
@dataclass(frozen=True)
class Exists(Formula):
    body: Abstraction

# Reduce(op::Abstraction, body::Abstraction, terms::Term[])
@dataclass(frozen=True)
class Reduce(Formula):
    op: Abstraction
    body: Abstraction
    terms: Sequence[Term]

# Conjunction(args::Formula[])
@dataclass(frozen=True)
class Conjunction(Formula):
    args: Sequence[Formula]

# Disjunction(args::Formula[])
@dataclass(frozen=True)
class Disjunction(Formula):
    args: Sequence[Formula]

# Not(arg::Formula)
@dataclass(frozen=True)
class Not(Formula):
    arg: Formula

# FFI(name::string, args::Abstraction[], terms::Term[])
@dataclass(frozen=True)
class FFI(Formula):
    name: str
    args: Sequence[Abstraction]
    terms: Sequence[Term]

# Atom(name::RelationId, terms::Term[])
@dataclass(frozen=True)
class Atom(Formula):
    name: RelationId
    terms: Sequence[Term]

# Pragma(name::string, terms::Term[])
@dataclass(frozen=True)
class Pragma(Formula):
    name: str
    terms: Sequence[Term]

# Primitive(name::string, terms::RelTerm[])
@dataclass(frozen=True)
class Primitive(Formula):
    name: str
    terms: Sequence[RelTerm]

# RelAtom(name::string, terms::RelTerm[])
@dataclass(frozen=True)
class RelAtom(Formula):
    name: str
    terms: Sequence[RelTerm]

# Cast(input::Term, result::Term)
@dataclass(frozen=True)
class Cast(Formula):
    input: Term
    result: Term

# Var(name::string)
@dataclass(frozen=True)
class Var(LqpNode):
    name: str

# UInt128Value(low::fixed64, high::fixed64)
@dataclass(frozen=True)
class UInt128Value(LqpNode):
    value: int

# Int128Value(low::fixed64, high::fixed64)
@dataclass(frozen=True)
class Int128Value(LqpNode):
    value: int

@dataclass(frozen=True)
class MissingValue(LqpNode):
    pass

# DateValue(year: int, month: int, day: int)
@dataclass(frozen=True)
class DateValue(LqpNode):
    value: dt.date

# DatetimeValue(year: int, month: int, day: int, hour: int, minute: int, second: int, microsecond: int)
@dataclass(frozen=True)
class DateTimeValue(LqpNode):
    value: dt.datetime

# DecimalValue(precision: int, scale: int, value: Decimal)
@dataclass(frozen=True)
class DecimalValue(LqpNode):
    precision: int
    scale: int
    value: Decimal

# BooleanValue(value: bool)
# Note: We need a custom BooleanValue class to distinguish it from Python's `int` type.
# Python's built-in `bool` is a subclass of `int`.
@dataclass(frozen=True)
class BooleanValue(LqpNode):
    value: bool

@dataclass(frozen=True)
class Value(LqpNode):
    value: Union[
        str,
        int,
        float,
        UInt128Value,
        Int128Value,
        MissingValue,
        DateValue,
        DateTimeValue,
        DecimalValue,
        BooleanValue
    ]

# SpecializedValue(value::Value)
@dataclass(frozen=True)
class SpecializedValue(LqpNode):
    value: Value

# Term := Var | Value
Term = Union[Var, Value]

# RelTerm := Term | SpecializedValue
RelTerm = Union[Term, SpecializedValue]

# Attribute(name::string, args::Constant[])
@dataclass(frozen=True)
class Attribute(LqpNode):
    name: str
    args: Sequence[Value]

# RelationId(id::UInt128)
@dataclass(frozen=True)
class RelationId(LqpNode):
    id: int
    def __post_init__(self):
        if self.id < 0 or self.id > 0xffffffffffffffffffffffffffffffff:
            raise ValueError("RelationId constructed with out of range (UInt128) number: {}".format(self.id))

    def __str__(self) -> str:
        if self.meta:
            return f"RelationId(meta={self.meta}, id={self.id})"
        return f"RelationId(id={self.id})"

    def __eq__(self, other) -> bool:
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

class TypeName(Enum):
    UNSPECIFIED = 0
    STRING = 1
    INT = 2
    FLOAT = 3
    UINT128 = 4
    INT128 = 5
    DATE = 6
    DATETIME = 7
    MISSING = 8
    DECIMAL = 9
    BOOLEAN = 10

    def __str__(self) -> str:
        return self.name

@dataclass(frozen=True)
class Type(LqpNode):
    type_name: TypeName
    parameters: Sequence[Value]

# --- Fragment Types ---

# FragmentId(id::bytes)
@dataclass(frozen=True)
class FragmentId(LqpNode):
    id: bytes

    def __eq__(self, other) -> bool:
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

# Fragment(id::FragmentId, declarations::Declaration[], debug_info::DebugInfo)
@dataclass(frozen=True)
class Fragment(LqpNode):
    id: FragmentId
    declarations: Sequence[Declaration]
    debug_info: DebugInfo

@dataclass(frozen=True)
class DebugInfo(LqpNode):
    id_to_orig_name: Dict[RelationId, str]

# --- Transaction Types ---

# Define(fragment::Fragment)
@dataclass(frozen=True)
class Define(LqpNode):
    fragment: Fragment

# Undefine(fragment_id::FragmentId)
@dataclass(frozen=True)
class Undefine(LqpNode):
    fragment_id: FragmentId

# Context(relations::RelationId[])
@dataclass(frozen=True)
class Context(LqpNode):
    relations: Sequence[RelationId]

# Write := Define | Undefine | Context
@dataclass(frozen=True)
class Write(LqpNode):
    write_type: Union[Define, Undefine, Context]

# Demand(relation_id::RelationId)
@dataclass(frozen=True)
class Demand(LqpNode):
    relation_id: RelationId

# Output(name::string?, relation_id::RelationId)
@dataclass(frozen=True)
class Output(LqpNode):
    name: Union[str, None]
    relation_id: RelationId

# ExportCSVConfig
@dataclass(frozen=True)
class ExportCSVConfig(LqpNode):
    path: str
    data_columns: Sequence[ExportCSVColumn]
    partition_size: Optional[int] = None
    compression: Optional[str] = None

    syntax_header_row: Optional[int] = None
    syntax_missing_string: Optional[str] = None
    syntax_delim: Optional[str] = None
    syntax_quotechar: Optional[str] = None
    syntax_escapechar: Optional[str] = None

@dataclass(frozen=True)
class ExportCSVColumn(LqpNode):
    column_name: str
    column_data: RelationId

# Export(name::string, relation_id::RelationId)
@dataclass(frozen=True)
class Export(LqpNode):
    # TODO: Once we add a JSON export, this should be union[ExportCSVConfig, ExportJSONConfig]
    config: ExportCSVConfig

# Abort(name::string?, relation_id::RelationId)
@dataclass(frozen=True)
class Abort(LqpNode):
    name: Union[str, None]
    relation_id: RelationId

# Read := Demand | Output | Export | WhatIf | Abort
@dataclass(frozen=True)
class Read(LqpNode):
    read_type: Union[Demand, Output, Export, WhatIf, Abort]

# Epoch(writes::Write[], reads::Read[])
@dataclass(frozen=True)
class Epoch(LqpNode):
    writes: Sequence[Write] = field(default_factory=list)
    reads: Sequence[Read] = field(default_factory=list)

# WhatIf(branch::string?, epoch::Epoch)
@dataclass(frozen=True)
class WhatIf(LqpNode):
    branch: Union[str, None]
    epoch: Epoch

# Transaction(epochs::Epoch[], configure::Configure)
@dataclass(frozen=True)
class Transaction(LqpNode):
    epochs: Sequence[Epoch]
    configure: Configure

# Configure(semantics_version::int, ivm_config::IVMConfig)
@dataclass(frozen=True)
class Configure(LqpNode):
    semantics_version: int
    ivm_config: IVMConfig

# IVMConfig(level::MaintenanceLevel)
@dataclass(frozen=True)
class IVMConfig(LqpNode):
    level: MaintenanceLevel

class MaintenanceLevel(Enum):
    UNSPECIFIED = 0
    OFF = 1
    AUTO = 2
    ALL = 3

    def __str__(self) -> str:
        return self.name



import lqp.ir as ir
import lqp.print as p
from typing import Any, Dict, List, Tuple, Sequence, Set, Union, Optional
from dataclasses import dataclass, is_dataclass, fields

class ValidationError(Exception):
    pass

class LqpVisitor:
    def visit(self, node: ir.LqpNode, *args: Any) -> None:
        method_name = f'visit_{node.__class__.__name__}'
        visitor_method = getattr(self, method_name, self.generic_visit)
        return visitor_method(node, *args)

    def generic_visit(self, node: ir.LqpNode, *args: Any) -> None:
        if not is_dataclass(node):
            raise ValidationError(f"Expected dataclass, got {type(node)}")
        for field in fields(node):
            value = getattr(node, field.name)
            if isinstance(value, ir.LqpNode):
                self.visit(value, *args)
            elif isinstance(value, (list, tuple)):
                for item in value:
                    if isinstance(item, ir.LqpNode):
                        self.visit(item, *args)
            elif isinstance(value, dict):
                for item in value.values():
                    if isinstance(item, ir.LqpNode):
                        self.visit(item, *args)

class UnusedVariableVisitor(LqpVisitor):
    def __init__(self, txn: ir.Transaction):
        self.scopes: List[Tuple[Set[str], Set[str]]] = []
        self.visit(txn)

    def _declare_var(self, var_name: str):
        if self.scopes:
            self.scopes[-1][0].add(var_name)

    def _mark_var_used(self, var: ir.Var):
        for declared, used in reversed(self.scopes):
            if var.name in declared:
                used.add(var.name)
                return
        raise ValidationError(f"Undeclared variable used at {var.meta}: '{var.name}'")

    def visit_Abstraction(self, node: ir.Abstraction):
        self.scopes.append((set(), set()))
        for var in node.vars:
            self._declare_var(var[0].name)
        self.visit(node.value)
        declared, used = self.scopes.pop()
        unused = declared - used
        if unused:
            for var_name in unused:
                # Allow an escape hatch for internal variables.
                if var_name.startswith("_"):
                    continue
                raise ValidationError(f"Unused variable declared: '{var_name}'")

    def visit_Var(self, node: ir.Var, *args: Any):
        self._mark_var_used(node)

# Checks for shadowing of variables. Raises ValidationError upon encountering such.
class ShadowedVariableFinder(LqpVisitor):
    def __init__(self, txn: ir.Transaction):
        self.visit(txn)

    # The varargs passed in must be a single set of strings.
    @staticmethod
    def args_ok(args: Sequence[Any]) -> bool:
        return (
            len(args) == 0 or
            (
                len(args) == 1 and
                isinstance(args[0], Set) and
                all(isinstance(s, str) for s in args[0])
            )
        )

    # Only Abstractions introduce variables.
    def visit_Abstraction(self, node: ir.Abstraction, *args: Any) -> None:
        assert ShadowedVariableFinder.args_ok(args)
        in_scope_names = set() if len(args) == 0 else args[0]

        for v in node.vars:
            var = v[0]
            if var.name in in_scope_names:
                raise ValidationError(f"Shadowed variable at {var.meta}: '{var.name}'")

        self.visit(node.value, in_scope_names | set(v[0].name for v in node.vars))

# Checks for invalid duplicate RelationIds. Duplicate relation IDs are only valid
# when they are within the same fragment in different epochs.
# Raises ValidationError upon encountering such.
class DuplicateRelationIdFinder(LqpVisitor):
    def __init__(self, txn: ir.Transaction):
        # RelationIds and where they have been defined. The integer represents
        # the epoch.
        self.seen_ids: Dict[ir.RelationId, Tuple[int, ir.FragmentId]] = dict()
        # We'll use this to give IDs to epochs as we visit them.
        self.curr_epoch: int = 0
        self.curr_fragment: Optional[ir.FragmentId] = None

        self.visit(txn)

    def visit_Def(self, node: ir.Def, *args: Any) -> None:
        assert self.curr_fragment is not None
        assert self.curr_epoch > 0

        if node.name in self.seen_ids:
            seen_in_epoch, seen_in_fragment = self.seen_ids[node.name]
            if self.curr_fragment != seen_in_fragment:
                # Dup ID, different fragments, same or different epoch.
                raise ValidationError(
                    f"Duplicate declaration across fragments at {node.meta}: '{node.name.id}'"
                )
            elif self.curr_epoch == seen_in_epoch:
                # Dup ID, same fragment, same epoch.
                raise ValidationError(
                    f"Duplicate declaration within fragment in epoch at {node.meta}: '{node.name.id}'"
                )
            # else: the final case (dup ID, same fragment, different epoch) is valid.

        self.seen_ids[node.name] = (self.curr_epoch, self.curr_fragment)

    def visit_Fragment(self, node: ir.Fragment, *args: Any) -> None:
        self.curr_fragment = node.id
        self.generic_visit(node, args)

    def visit_Epoch(self, node: ir.Epoch, *args: Any) -> None:
        self.curr_epoch += 1
        self.generic_visit(node, args)

    def visit_Algorithm(self, node: ir.Algorithm, *args: Any) -> None:
        # Only the Defs in init are globally visible so don't visit body Defs.
        for d in node.global_:
            if d in self.seen_ids:
                raise ValidationError(
                    f"Duplicate declaration at {d.meta}: '{d.id}'"
                )
            else:
                assert self.curr_fragment is not None
                self.seen_ids[d] = (self.curr_epoch, self.curr_fragment)

# Checks that Instructions are applied to the correct number and types of terms.
# Assumes UnusedVariableVisitor has passed.
class AtomTypeChecker(LqpVisitor):
    Instructions = Union[ir.Def, ir.Assign, ir.Break, ir.Upsert]
    # Helper to get all Defs defined in a Transaction. We are only interested
    # in globally visible Defs thus ignore Loop bodies.
    @staticmethod
    def collect_global_defs(txn: ir.Transaction) -> List[Instructions]:
        # Visitor to do the work.
        class DefCollector(LqpVisitor):
            def __init__(self, txn: ir.Transaction):
                self.atoms: List[AtomTypeChecker.Instructions] = []
                self.visit(txn)

            def visit_Def(self, node: ir.Def) -> None:
                self.atoms.append(node)

            def visit_Algorithm(self, node:ir.Algorithm):
                self.atoms.extend([d for d in node.body.constructs if isinstance(d, AtomTypeChecker.Instructions)])

            def visit_Loop(self, node: ir.Loop) -> None:
                self.atoms.extend([d for d in node.init if isinstance(d, AtomTypeChecker.Instructions)])
                # Don't touch the body, they are not globally visible. Treat
                # this node as a leaf.
        return DefCollector(txn).atoms

    # Helper to map Constants to their TypeName.
    @staticmethod
    def constant_type(c: ir.Value) -> ir.TypeName: # type: ignore
        if isinstance(c.value, str):
            return ir.TypeName.STRING
        elif isinstance(c.value, int):
            return ir.TypeName.INT
        elif isinstance(c.value, float):
            return ir.TypeName.FLOAT
        elif isinstance(c.value, ir.Int128Value):
            return ir.TypeName.INT128
        elif isinstance(c.value, ir.UInt128Value):
            return ir.TypeName.UINT128
        elif isinstance(c.value, ir.MissingValue):
            return ir.TypeName.MISSING
        elif isinstance(c.value, ir.DecimalValue):
            return ir.TypeName.DECIMAL
        elif isinstance(c.value, ir.DateValue):
            return ir.TypeName.DATE
        elif isinstance(c.value, ir.DateTimeValue):
            return ir.TypeName.DATETIME
        elif isinstance(c.value, ir.BooleanValue):
            return ir.TypeName.BOOLEAN
        else:
            assert False, f"Unknown constant type: {type(c.value)}"

    @staticmethod
    def type_error_message(atom: ir.Atom, index: int, expected: ir.TypeName, actual: ir.TypeName) -> str:
        term = atom.terms[index]
        pretty_term = p.to_str(term, 0)
        return \
            f"Incorrect type for '{atom.name.id}' atom at index {index} ('{pretty_term}') at {atom.meta}: " +\
            f"expected {expected} term, got {actual}"

    # Return a list of the types of the parameters of a Def.
    @staticmethod
    def get_relation_sig(d: Instructions):
        # v[1] holds the TypeName.
        return [v[1].type_name for v in d.body.vars]

    # The varargs passed be a State or nothing at all.
    @staticmethod
    def args_ok(args: tuple[Any]) -> bool:
        return len(args) == 1 and isinstance(args[0], AtomTypeChecker.State)

    # What we pass around to the visit methods.
    @dataclass(frozen=True)
    class State:
        # Maps relations in scope to their types.
        relation_types: Dict[ir.RelationId, List[ir.TypeName]]
        # Maps variables in scope to their type.
        var_types: Dict[str, ir.TypeName]

    def __init__(self, txn: ir.Transaction):
        state = AtomTypeChecker.State(
            {
                d.name : AtomTypeChecker.get_relation_sig(d)
                for d in AtomTypeChecker.collect_global_defs(txn)
            },
            # No variables declared yet.
            {},
        )
        self.visit(txn, state)

    # Visit Abstractions to collect the types of variables.
    def visit_Abstraction(self, node: ir.Abstraction, *args: Any) -> None:
        assert AtomTypeChecker.args_ok(args)
        state = args[0]

        self.generic_visit(
            node,
            AtomTypeChecker.State(
                state.relation_types,
                state.var_types | {v.name : t.type_name for (v, t) in node.vars},
            ),
        )

    # Visit Loops as body Defs are not global and need to be introduced to their
    # children.
    def visit_Loop(self, node: ir.Loop, *args: Any) -> None:
        assert AtomTypeChecker.args_ok(args)
        state = args[0]

        for d in node.init:
            self.visit(d, state)

        for decl in node.body.constructs:
            if isinstance(decl, ir.Instruction):
                self.visit(
                    decl,
                    AtomTypeChecker.State(
                        {decl.name : AtomTypeChecker.get_relation_sig(decl)} | state.relation_types, #type: ignore
                        state.var_types,
                    ),
                )
            else:
                self.visit(decl, state)

    def visit_Atom(self, node: ir.Atom, *args: Any) -> None:
        assert AtomTypeChecker.args_ok(args)
        state = args[0]

        # Relation may have been defined in another transaction, we don't know,
        # so ignore this atom.
        if node.name in state.relation_types:
            relation_type_sig = state.relation_types[node.name]

            # Check arity.
            atom_arity = len(node.terms)
            relation_arity = len(relation_type_sig)
            if atom_arity != relation_arity:
                raise ValidationError(
                    f"Incorrect arity for '{node.name.id}' atom at {node.meta}: " +\
                    f"expected {relation_arity} term{'' if relation_arity == 1 else 's'}, got {atom_arity}"
                )

            # Check types.
            for (i, (term, relation_type)) in enumerate(zip(node.terms, relation_type_sig)):
                # var_types[term] is okay because we assume UnusedVariableVisitor.
                term_type = state.var_types[term.name] if isinstance(term, ir.Var) else AtomTypeChecker.constant_type(term)
                if term_type.value != relation_type.value:
                    raise ValidationError(
                        AtomTypeChecker.type_error_message(node, i, relation_type, term_type)
                    )

        # This is a leaf for our purposes, no need to recurse further.

# Checks for the definition (Define) of duplicate Fragment(Ids) within an Epoch.
# Raises ValidationError upon encountering such.
class DuplicateFragmentDefinitionFinder(LqpVisitor):
    def __init__(self, txn: ir.Transaction):
        # Instead of passing this back and forth, we are going to clear this
        # when we visit an Epoch and let it fill, checking for duplicates
        # when we visit descendent Fragments. When we visit another Epoch, it'll
        # be cleared again, etc.
        self.seen_ids: Set[ir.FragmentId] = set()
        self.visit(txn)

    def visit_Epoch(self, node: ir.Epoch, *args: Any) -> None:
        self.seen_ids.clear()
        self.generic_visit(node)

    # We could visit_Fragment instead (no node has a Fragment child except
    # Define) but the point of this pass is to find duplicate Fragments
    # being _defined_ so this is a bit more fitting.
    def visit_Define(self, node: ir.Define, *args: Any) -> None:
        if node.fragment.id in self.seen_ids:
            id_str = node.fragment.id.id.decode("utf-8")
            raise ValidationError(
                f"Duplicate fragment within an epoch at {node.meta}: '{id_str}'"
            )
        else:
            self.seen_ids.add(node.fragment.id)

        # No need to recurse further; no descendent Epochs/Fragments.


# Loopy contract: Break rules can only go in inits
class LoopyBadBreakFinder(LqpVisitor):
    def __init__(self, txn: ir.Transaction):
        self.visit(txn)

    def visit_Loop(self, node: ir.Loop, *args: Any) -> None:
        for i in node.init:
            if isinstance(i, ir.Break):
                raise ValidationError(
                    f"Break rule found outside of body at {i.meta}: '{i.name.id}'"
                )

# Loopy contract: Algorithm globals cannot be in loop body unless they were already in init
class LoopyBadGlobalFinder(LqpVisitor):
    def __init__(self, txn: ir.Transaction):
        self.globals: Set[ir.RelationId] = set()
        self.init: Set[ir.RelationId] = set()
        self.visit(txn)

    def visit_Algorithm(self, node: ir.Algorithm, *args: Any) -> None:
        self.globals = self.globals.union(node.global_)
        self.visit(node.body)
        self.globals.clear()

    def visit_Loop(self, node: ir.Loop, *args: Any) -> None:
        self.init = {x.name for x in node.init if isinstance(x, (ir.Break, ir.Assign, ir.Upsert))}
        for i in node.body.constructs:
            if isinstance(i, (ir.Break, ir.Assign, ir.Upsert)):
                if (i.name in self.globals) and (i.name not in self.init):
                    raise ValidationError(
                        f"Global rule found in body at {i.meta}: '{i.name.id}'"
                    )

class LoopyUpdatesShouldBeAtoms(LqpVisitor):
    def __init__(self, txn: ir.Transaction):
        self.generic_visit(txn)

    def visit_instruction_with_atom_body(self, node: Any, *args: Any) -> None:
        if not isinstance(node.body.value, ir.Atom):
            instruction_type = node.__class__.__name__
            raise ValidationError(f"{instruction_type} at {node.meta} must have an Atom as its value")

    visit_Copy = visit_MonoidDef = visit_MonusDef = visit_Upsert = visit_instruction_with_atom_body

def validate_lqp(lqp: ir.Transaction):
    ShadowedVariableFinder(lqp)
    UnusedVariableVisitor(lqp)
    DuplicateRelationIdFinder(lqp)
    DuplicateFragmentDefinitionFinder(lqp)
    AtomTypeChecker(lqp)
    LoopyBadBreakFinder(lqp)
    LoopyBadGlobalFinder(lqp)
    LoopyUpdatesShouldBeAtoms(lqp)



import pytest
from lqp.ir import RelationId, SourceInfo

def test_relation_id_str_no_meta():
    rel_id = RelationId(meta=None, id=12345)
    assert str(rel_id) == "RelationId(id=12345)"

def test_relation_id_str_with_meta():
    source_info = SourceInfo(file="test.lqp", line=1, column=10)
    rel_id = RelationId(meta=source_info, id=67890)
    assert str(rel_id) == "RelationId(meta=test.lqp:1:10, id=67890)"



from lqp.proto.v1 import fragments_pb2 as _fragments_pb2
from lqp.proto.v1 import logic_pb2 as _logic_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MaintenanceLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    MAINTENANCE_LEVEL_UNSPECIFIED: _ClassVar[MaintenanceLevel]
    MAINTENANCE_LEVEL_OFF: _ClassVar[MaintenanceLevel]
    MAINTENANCE_LEVEL_AUTO: _ClassVar[MaintenanceLevel]
    MAINTENANCE_LEVEL_ALL: _ClassVar[MaintenanceLevel]
MAINTENANCE_LEVEL_UNSPECIFIED: MaintenanceLevel
MAINTENANCE_LEVEL_OFF: MaintenanceLevel
MAINTENANCE_LEVEL_AUTO: MaintenanceLevel
MAINTENANCE_LEVEL_ALL: MaintenanceLevel

class Transaction(_message.Message):
    __slots__ = ("epochs", "configure")
    EPOCHS_FIELD_NUMBER: _ClassVar[int]
    CONFIGURE_FIELD_NUMBER: _ClassVar[int]
    epochs: _containers.RepeatedCompositeFieldContainer[Epoch]
    configure: Configure
    def __init__(self, epochs: _Optional[_Iterable[_Union[Epoch, _Mapping]]] = ..., configure: _Optional[_Union[Configure, _Mapping]] = ...) -> None: ...

class Configure(_message.Message):
    __slots__ = ("semantics_version", "ivm_config")
    SEMANTICS_VERSION_FIELD_NUMBER: _ClassVar[int]
    IVM_CONFIG_FIELD_NUMBER: _ClassVar[int]
    semantics_version: int
    ivm_config: IVMConfig
    def __init__(self, semantics_version: _Optional[int] = ..., ivm_config: _Optional[_Union[IVMConfig, _Mapping]] = ...) -> None: ...

class IVMConfig(_message.Message):
    __slots__ = ("level",)
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    level: MaintenanceLevel
    def __init__(self, level: _Optional[_Union[MaintenanceLevel, str]] = ...) -> None: ...

class Epoch(_message.Message):
    __slots__ = ("writes", "reads")
    WRITES_FIELD_NUMBER: _ClassVar[int]
    READS_FIELD_NUMBER: _ClassVar[int]
    writes: _containers.RepeatedCompositeFieldContainer[Write]
    reads: _containers.RepeatedCompositeFieldContainer[Read]
    def __init__(self, writes: _Optional[_Iterable[_Union[Write, _Mapping]]] = ..., reads: _Optional[_Iterable[_Union[Read, _Mapping]]] = ...) -> None: ...

class Write(_message.Message):
    __slots__ = ("define", "undefine", "context")
    DEFINE_FIELD_NUMBER: _ClassVar[int]
    UNDEFINE_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    define: Define
    undefine: Undefine
    context: Context
    def __init__(self, define: _Optional[_Union[Define, _Mapping]] = ..., undefine: _Optional[_Union[Undefine, _Mapping]] = ..., context: _Optional[_Union[Context, _Mapping]] = ...) -> None: ...

class Define(_message.Message):
    __slots__ = ("fragment",)
    FRAGMENT_FIELD_NUMBER: _ClassVar[int]
    fragment: _fragments_pb2.Fragment
    def __init__(self, fragment: _Optional[_Union[_fragments_pb2.Fragment, _Mapping]] = ...) -> None: ...

class Undefine(_message.Message):
    __slots__ = ("fragment_id",)
    FRAGMENT_ID_FIELD_NUMBER: _ClassVar[int]
    fragment_id: _fragments_pb2.FragmentId
    def __init__(self, fragment_id: _Optional[_Union[_fragments_pb2.FragmentId, _Mapping]] = ...) -> None: ...

class Context(_message.Message):
    __slots__ = ("relations",)
    RELATIONS_FIELD_NUMBER: _ClassVar[int]
    relations: _containers.RepeatedCompositeFieldContainer[_logic_pb2.RelationId]
    def __init__(self, relations: _Optional[_Iterable[_Union[_logic_pb2.RelationId, _Mapping]]] = ...) -> None: ...

class ExportCSVConfig(_message.Message):
    __slots__ = ("path", "data_columns", "partition_size", "compression", "syntax_header_row", "syntax_missing_string", "syntax_delim", "syntax_quotechar", "syntax_escapechar")
    PATH_FIELD_NUMBER: _ClassVar[int]
    DATA_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    PARTITION_SIZE_FIELD_NUMBER: _ClassVar[int]
    COMPRESSION_FIELD_NUMBER: _ClassVar[int]
    SYNTAX_HEADER_ROW_FIELD_NUMBER: _ClassVar[int]
    SYNTAX_MISSING_STRING_FIELD_NUMBER: _ClassVar[int]
    SYNTAX_DELIM_FIELD_NUMBER: _ClassVar[int]
    SYNTAX_QUOTECHAR_FIELD_NUMBER: _ClassVar[int]
    SYNTAX_ESCAPECHAR_FIELD_NUMBER: _ClassVar[int]
    path: str
    data_columns: _containers.RepeatedCompositeFieldContainer[ExportCSVColumn]
    partition_size: int
    compression: str
    syntax_header_row: bool
    syntax_missing_string: str
    syntax_delim: str
    syntax_quotechar: str
    syntax_escapechar: str
    def __init__(self, path: _Optional[str] = ..., data_columns: _Optional[_Iterable[_Union[ExportCSVColumn, _Mapping]]] = ..., partition_size: _Optional[int] = ..., compression: _Optional[str] = ..., syntax_header_row: bool = ..., syntax_missing_string: _Optional[str] = ..., syntax_delim: _Optional[str] = ..., syntax_quotechar: _Optional[str] = ..., syntax_escapechar: _Optional[str] = ...) -> None: ...

class ExportCSVColumn(_message.Message):
    __slots__ = ("column_name", "column_data")
    COLUMN_NAME_FIELD_NUMBER: _ClassVar[int]
    COLUMN_DATA_FIELD_NUMBER: _ClassVar[int]
    column_name: str
    column_data: _logic_pb2.RelationId
    def __init__(self, column_name: _Optional[str] = ..., column_data: _Optional[_Union[_logic_pb2.RelationId, _Mapping]] = ...) -> None: ...

class Read(_message.Message):
    __slots__ = ("demand", "output", "what_if", "abort", "export")
    DEMAND_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_FIELD_NUMBER: _ClassVar[int]
    WHAT_IF_FIELD_NUMBER: _ClassVar[int]
    ABORT_FIELD_NUMBER: _ClassVar[int]
    EXPORT_FIELD_NUMBER: _ClassVar[int]
    demand: Demand
    output: Output
    what_if: WhatIf
    abort: Abort
    export: Export
    def __init__(self, demand: _Optional[_Union[Demand, _Mapping]] = ..., output: _Optional[_Union[Output, _Mapping]] = ..., what_if: _Optional[_Union[WhatIf, _Mapping]] = ..., abort: _Optional[_Union[Abort, _Mapping]] = ..., export: _Optional[_Union[Export, _Mapping]] = ...) -> None: ...

class Demand(_message.Message):
    __slots__ = ("relation_id",)
    RELATION_ID_FIELD_NUMBER: _ClassVar[int]
    relation_id: _logic_pb2.RelationId
    def __init__(self, relation_id: _Optional[_Union[_logic_pb2.RelationId, _Mapping]] = ...) -> None: ...

class Output(_message.Message):
    __slots__ = ("name", "relation_id")
    NAME_FIELD_NUMBER: _ClassVar[int]
    RELATION_ID_FIELD_NUMBER: _ClassVar[int]
    name: str
    relation_id: _logic_pb2.RelationId
    def __init__(self, name: _Optional[str] = ..., relation_id: _Optional[_Union[_logic_pb2.RelationId, _Mapping]] = ...) -> None: ...

class Export(_message.Message):
    __slots__ = ("csv_config",)
    CSV_CONFIG_FIELD_NUMBER: _ClassVar[int]
    csv_config: ExportCSVConfig
    def __init__(self, csv_config: _Optional[_Union[ExportCSVConfig, _Mapping]] = ...) -> None: ...

class WhatIf(_message.Message):
    __slots__ = ("branch", "epoch")
    BRANCH_FIELD_NUMBER: _ClassVar[int]
    EPOCH_FIELD_NUMBER: _ClassVar[int]
    branch: str
    epoch: Epoch
    def __init__(self, branch: _Optional[str] = ..., epoch: _Optional[_Union[Epoch, _Mapping]] = ...) -> None: ...

class Abort(_message.Message):
    __slots__ = ("name", "relation_id")
    NAME_FIELD_NUMBER: _ClassVar[int]
    RELATION_ID_FIELD_NUMBER: _ClassVar[int]
    name: str
    relation_id: _logic_pb2.RelationId
    def __init__(self, name: _Optional[str] = ..., relation_id: _Optional[_Union[_logic_pb2.RelationId, _Mapping]] = ...) -> None: ...



from lqp.proto.v1 import logic_pb2 as _logic_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Fragment(_message.Message):
    __slots__ = ("id", "declarations", "debug_info")
    ID_FIELD_NUMBER: _ClassVar[int]
    DECLARATIONS_FIELD_NUMBER: _ClassVar[int]
    DEBUG_INFO_FIELD_NUMBER: _ClassVar[int]
    id: FragmentId
    declarations: _containers.RepeatedCompositeFieldContainer[_logic_pb2.Declaration]
    debug_info: DebugInfo
    def __init__(self, id: _Optional[_Union[FragmentId, _Mapping]] = ..., declarations: _Optional[_Iterable[_Union[_logic_pb2.Declaration, _Mapping]]] = ..., debug_info: _Optional[_Union[DebugInfo, _Mapping]] = ...) -> None: ...

class DebugInfo(_message.Message):
    __slots__ = ("ids", "orig_names")
    IDS_FIELD_NUMBER: _ClassVar[int]
    ORIG_NAMES_FIELD_NUMBER: _ClassVar[int]
    ids: _containers.RepeatedCompositeFieldContainer[_logic_pb2.RelationId]
    orig_names: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, ids: _Optional[_Iterable[_Union[_logic_pb2.RelationId, _Mapping]]] = ..., orig_names: _Optional[_Iterable[str]] = ...) -> None: ...

class FragmentId(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: bytes
    def __init__(self, id: _Optional[bytes] = ...) -> None: ...



import lqp.ir as ir
from lqp.proto.v1 import logic_pb2, fragments_pb2, transactions_pb2
from typing import Union, Dict, Any
from functools import reduce

# Maps ir.TypeNames to the associated Proto message for *non-paremetric types*. Used to generically construct non-parametric types.
# Parametric types should be handled in convert_type
non_parametric_types = {
    ir.TypeName.UNSPECIFIED: "UnspecifiedType",
    ir.TypeName.STRING: "StringType",
    ir.TypeName.INT: "IntType",
    ir.TypeName.FLOAT: "FloatType",
    ir.TypeName.UINT128: "UInt128Type",
    ir.TypeName.INT128: "Int128Type",
    ir.TypeName.DATE: "DateType",
    ir.TypeName.DATETIME: "DateTimeType",
    ir.TypeName.MISSING: "MissingType",
    ir.TypeName.BOOLEAN: "BooleanType",
}

def convert_type(rt: ir.Type) -> logic_pb2.Type:
    if rt.type_name == ir.TypeName.DECIMAL :
        assert isinstance(rt.parameters[0].value, int) and isinstance(rt.parameters[1].value, int), "DECIMAL parameters are not integers"
        assert len(rt.parameters) == 2, f"DECIMAL parameters should have only precision and scale, got {len(rt.parameters)} arguments"
        assert rt.parameters[0].value <= 38, f"DECIMAL precision must be less than 38, got {rt.parameters[0]}"
        assert rt.parameters[1].value <= rt.parameters[0].value, f"DECIMAL precision ({rt.parameters[0]}) must be at least scale ({rt.parameters[1]})"
        return logic_pb2.Type(
            decimal_type=logic_pb2.DecimalType(
                precision=rt.parameters[0].value, scale=rt.parameters[1].value
            )
        )
    else:
        cls = getattr(logic_pb2, non_parametric_types[rt.type_name])
        return logic_pb2.Type(**{str(rt.type_name).lower()+"_type": cls()})

def convert_uint128(val: ir.UInt128Value) -> logic_pb2.UInt128Value:
    low = val.value & 0xFFFFFFFFFFFFFFFF
    high = (val.value >> 64) & 0xFFFFFFFFFFFFFFFF
    return logic_pb2.UInt128Value(low=low, high=high)

def convert_int128(val: ir.Int128Value) -> logic_pb2.Int128Value:
    low = val.value & 0xFFFFFFFFFFFFFFFF
    high = (val.value >> 64) & 0xFFFFFFFFFFFFFFFF
    return logic_pb2.Int128Value(low=low, high=high)

def convert_date(val: ir.DateValue) -> logic_pb2.DateValue:
    return logic_pb2.DateValue(year=val.value.year, month=val.value.month, day=val.value.day)

def convert_datetime(val: ir.DateTimeValue) -> logic_pb2.DateTimeValue:
    return logic_pb2.DateTimeValue(
        year=val.value.year,
        month=val.value.month,
        day=val.value.day,
        hour=val.value.hour,
        minute=val.value.minute,
        second=val.value.second,
        microsecond=val.value.microsecond
    )

def convert_decimal(val: ir.DecimalValue) -> logic_pb2.DecimalValue:
    _, digits, exponent = val.value.as_tuple()
    value = reduce(lambda rst, d: rst * 10 + d, digits)

    assert isinstance(exponent, int)
    assert isinstance(value, int)

    # Adjust value by the exponent. Python's decimal values are (sign, digits, exponent),
    # so if we have digits 12300 with exponent -4, but we need `scale` of 6, then we need to
    # multiply the digits by 10 ** 2 (i.e., 10 ** (6 + -4)) to get the physical value of
    # 1230000.
    # Ensure we stay in the integer realm when the exponent outweighs the scale, e.g.
    # value = 4.4000000000000003552713678800500929355621337890625
    modifier = val.scale + exponent
    if modifier >= 0:
        value *= 10 ** modifier
    else:
        value //= 10 ** (-modifier)
    value = ir.Int128Value(value=value, meta=val.meta)

    return logic_pb2.DecimalValue(
        precision=val.precision,
        scale=val.scale,
        value=convert_int128(value),
    )

def convert_value(pv: ir.Value) -> logic_pb2.Value:
    if isinstance(pv.value, str):
        return logic_pb2.Value(string_value=pv.value)
    elif isinstance(pv.value, ir.MissingValue):
        return logic_pb2.Value(missing_value=logic_pb2.MissingValue())
    elif isinstance(pv.value, int):
        assert pv.value.bit_length() <= 64, "Integer value exceeds 64 bits"
        return logic_pb2.Value(int_value=pv.value)
    elif isinstance(pv.value, float):
        return logic_pb2.Value(float_value=pv.value)
    elif isinstance(pv.value, ir.UInt128Value):
        return logic_pb2.Value(
            uint128_value=convert_uint128(pv.value)
        )
    elif isinstance(pv.value, ir.Int128Value):
        return logic_pb2.Value(
            int128_value=convert_int128(pv.value)
        )
    elif isinstance(pv.value, ir.DateValue):
        return logic_pb2.Value(
            date_value=convert_date(pv.value)
        )
    elif isinstance(pv.value, ir.DateTimeValue):
        return logic_pb2.Value(
            datetime_value=convert_datetime(pv.value)
        )
    elif isinstance(pv.value, ir.DecimalValue):
        return logic_pb2.Value(
            decimal_value=convert_decimal(pv.value)
        )
    elif isinstance(pv.value, ir.BooleanValue):
        return logic_pb2.Value(boolean_value=pv.value.value)
    else:
        raise TypeError(f"Unsupported Value type: {type(pv.value)}")

def convert_var(v: ir.Var) -> logic_pb2.Var:
    return logic_pb2.Var(name=v.name)

def convert_term(t: ir.Term) -> logic_pb2.Term:
    if isinstance(t, ir.Var):
        return logic_pb2.Term(var=convert_var(t))
    else:
        return logic_pb2.Term(constant=convert_value(t))

def convert_relterm(t: ir.RelTerm) -> logic_pb2.RelTerm:
    if isinstance(t, ir.SpecializedValue):
        return logic_pb2.RelTerm(specialized_value=convert_value(t.value))
    else:
        return logic_pb2.RelTerm(term=convert_term(t))

def convert_relation_id(rid: ir.RelationId) -> logic_pb2.RelationId:
    id_low = rid.id & 0xFFFFFFFFFFFFFFFF
    id_high = (rid.id >> 64) & 0xFFFFFFFFFFFFFFFF
    return logic_pb2.RelationId(id_low=id_low, id_high=id_high)

def convert_fragment_id(fid: ir.FragmentId) -> fragments_pb2.FragmentId:
    return fragments_pb2.FragmentId(id=fid.id)

def convert_attribute(attr: ir.Attribute) -> logic_pb2.Attribute:
    return logic_pb2.Attribute(
        name=attr.name,
        args=[convert_value(arg) for arg in attr.args]
)

def convert_abstraction(abst: ir.Abstraction) -> logic_pb2.Abstraction:
    bindings = [logic_pb2.Binding(var=convert_var(var_tuple[0]), type=convert_type(var_tuple[1]))
                for var_tuple in abst.vars]
    return logic_pb2.Abstraction(
        vars=bindings,
        value=convert_formula(abst.value)
    )

def convert_formula(f: ir.Formula) -> logic_pb2.Formula:
    if isinstance(f, ir.Exists):
        return logic_pb2.Formula(exists=logic_pb2.Exists(body=convert_abstraction(f.body)))
    elif isinstance(f, ir.Reduce):
        return logic_pb2.Formula(reduce=logic_pb2.Reduce(
            op=convert_abstraction(f.op),
            body=convert_abstraction(f.body),
            terms=[convert_term(t) for t in f.terms]
        ))
    elif isinstance(f, ir.Conjunction):
        return logic_pb2.Formula(conjunction=logic_pb2.Conjunction(args=[convert_formula(arg) for arg in f.args]))
    elif isinstance(f, ir.Disjunction):
        return logic_pb2.Formula(disjunction=logic_pb2.Disjunction(args=[convert_formula(arg) for arg in f.args]))
    elif isinstance(f, ir.Not):
        return logic_pb2.Formula(**{"not": logic_pb2.Not(arg=convert_formula(f.arg))}) # type: ignore
    elif isinstance(f, ir.FFI):
        return logic_pb2.Formula(ffi=logic_pb2.FFI(
            name=f.name,
            args=[convert_abstraction(arg) for arg in f.args],
            terms=[convert_term(t) for t in f.terms]
        ))
    elif isinstance(f, ir.Atom):
        return logic_pb2.Formula(atom=logic_pb2.Atom(
            name=convert_relation_id(f.name),
            terms=[convert_term(t) for t in f.terms]
        ))
    elif isinstance(f, ir.Pragma):
        return logic_pb2.Formula(pragma=logic_pb2.Pragma(
            name=f.name,
            terms=[convert_term(t) for t in f.terms]
        ))
    elif isinstance(f, ir.Primitive):
        primitive_proto = logic_pb2.Primitive(name=f.name, terms=[convert_relterm(t) for t in f.terms])
        return logic_pb2.Formula(primitive=primitive_proto)
    elif isinstance(f, ir.RelAtom):
        rel_atom_proto = logic_pb2.RelAtom(name=f.name, terms=[convert_relterm(t) for t in f.terms])
        return logic_pb2.Formula(rel_atom=rel_atom_proto)
    elif isinstance(f, ir.Cast):
        return logic_pb2.Formula(cast=logic_pb2.Cast(
            input=convert_term(f.input),
            result=convert_term(f.result)
        ))
    else:
        raise TypeError(f"Unsupported Formula type: {type(f)}")

def convert_def(d: ir.Def) -> logic_pb2.Def:
    return logic_pb2.Def(
        name=convert_relation_id(d.name),
        body=convert_abstraction(d.body),
        attrs=[convert_attribute(attr) for attr in d.attrs]
    )

def convert_loop(l: ir.Loop) -> logic_pb2.Loop:
    return logic_pb2.Loop(
        init=[convert_instruction(init_def) for init_def in l.init],
        body=convert_script(l.body)
    )

def convert_declaration(decl: ir.Declaration) -> logic_pb2.Declaration:
    from typing import Dict, Any
    if isinstance(decl, ir.Def):
        decl_dict: Dict[str, Any] = {'def': convert_def(decl)}
        return logic_pb2.Declaration(**decl_dict)  # type: ignore
    elif isinstance(decl, ir.Algorithm):
        algorithm_dict: Dict[str, Any] = {'algorithm': convert_algorithm(decl)}
        return logic_pb2.Declaration(**algorithm_dict)
    else:
        raise TypeError(f"Unsupported Declaration type: {type(decl)}")

def convert_algorithm(algo: ir.Algorithm)-> logic_pb2.Algorithm:
    dict: Dict[str, Any] = {
        'global': [convert_relation_id(id) for id in algo.global_],
        'body':convert_script(algo.body)
    }
    return logic_pb2.Algorithm(**dict)

def convert_instruction(instr: ir.Instruction) -> logic_pb2.Instruction:
    from typing import Dict, Any
    if isinstance(instr, ir.Assign):
        dict: Dict[str, Any] = {'assign': convert_assign(instr)}
        return logic_pb2.Instruction(**dict)
    elif isinstance(instr, ir.Break):
        dict: Dict[str, Any] = {'break': convert_break(instr)}
        return logic_pb2.Instruction(**dict)
    elif isinstance(instr, ir.Upsert):
        dict: Dict[str, Any] = {'upsert': convert_upsert(instr)}
        return logic_pb2.Instruction(**dict)
    elif isinstance(instr, ir.Copy):
        dict: Dict[str, Any] = {'copy': convert_copy(instr)}
        return logic_pb2.Instruction(**dict)
    elif isinstance(instr, ir.MonoidDef):
        dict: Dict[str, Any] = {'monoid_def': convert_monoid_def(instr)}
        return logic_pb2.Instruction(**dict)
    elif isinstance(instr, ir.MonusDef):
        dict: Dict[str, Any] = {'monus_def': convert_monus_def(instr)}
        return logic_pb2.Instruction(**dict)
    else:
        raise TypeError(f"Unsupported Instruction type: {type(instr)}")

def convert_assign(instr: ir.Assign) -> logic_pb2.Assign:
    return logic_pb2.Assign(
        name=convert_relation_id(instr.name),
        body=convert_abstraction(instr.body),
        attrs=[convert_attribute(attr) for attr in instr.attrs]
    )
def convert_break(instr: ir.Break) -> logic_pb2.Break:
    return logic_pb2.Break(
        name=convert_relation_id(instr.name),
        body=convert_abstraction(instr.body),
        attrs=[convert_attribute(attr) for attr in instr.attrs]
    )
def convert_upsert(instr: ir.Upsert) -> logic_pb2.Upsert:
    return logic_pb2.Upsert(
        name=convert_relation_id(instr.name),
        body=convert_abstraction(instr.body),
        attrs=[convert_attribute(attr) for attr in instr.attrs]
    )
def convert_copy(instr: ir.Copy) -> logic_pb2.Copy:
    return logic_pb2.Copy(
        name=convert_relation_id(instr.name),
        body=convert_abstraction(instr.body),
        attrs=[convert_attribute(attr) for attr in instr.attrs]
    )
def convert_monoid_def(instr: ir.MonoidDef) -> logic_pb2.MonoidDef:
    return logic_pb2.MonoidDef(
        monoid=convert_monoid(instr.monoid),
        name=convert_relation_id(instr.name),
        body=convert_abstraction(instr.body),
        attrs=[convert_attribute(attr) for attr in instr.attrs]
    )
def convert_monus_def(instr: ir.MonusDef) -> logic_pb2.MonusDef:
    return logic_pb2.MonusDef(
        monoid=convert_monoid(instr.monoid),
        name=convert_relation_id(instr.name),
        body=convert_abstraction(instr.body),
        attrs=[convert_attribute(attr) for attr in instr.attrs]
    )
def convert_monoid(monoid: ir.Monoid) -> logic_pb2.Monoid:
    from typing import Dict, Any
    if isinstance(monoid, ir.OrMonoid):
        return logic_pb2.Monoid(**{'or_monoid': logic_pb2.OrMonoid()})  # type: ignore
    elif isinstance(monoid, ir.SumMonoid):
        type = convert_type(monoid.type)
        return logic_pb2.Monoid(**{'sum_monoid': logic_pb2.SumMonoid(type=type)})  # type: ignore
    elif isinstance(monoid, ir.MinMonoid):
        type = convert_type(monoid.type)
        return logic_pb2.Monoid(**{'min_monoid': logic_pb2.MinMonoid(type=type)})  # type: ignore
    elif isinstance(monoid, ir.MaxMonoid):
        type = convert_type(monoid.type)
        return logic_pb2.Monoid(**{'max_monoid': logic_pb2.MaxMonoid(type=type)})  # type: ignore
    else:
        raise TypeError(f"Unsupported Monoid: {monoid}")


def convert_script(script: ir.Script) -> logic_pb2.Script:
    return logic_pb2.Script(constructs=[convert_construct(c) for c in script.constructs])

def convert_construct(construct: ir.Construct) -> logic_pb2.Construct:
    from typing import Dict, Any
    if isinstance(construct, ir.Loop):
        loop_dict: Dict[str, Any] = {'loop': convert_loop(construct)}
        return logic_pb2.Construct(**loop_dict)  # type: ignore
    elif isinstance(construct, ir.Instruction):
        instruction_dict: Dict[str, Any] = {'instruction': convert_instruction(construct)}
        return logic_pb2.Construct(**instruction_dict)
    else:
        raise TypeError(f"Unsupported Construct type: {type(construct)}")

def convert_fragment(frag: ir.Fragment) -> fragments_pb2.Fragment:
    return fragments_pb2.Fragment(
        id=convert_fragment_id(frag.id),
        declarations=[convert_declaration(decl) for decl in frag.declarations],
        debug_info=convert_debug_info(frag.debug_info)
    )

def convert_debug_info(info: ir.DebugInfo) -> fragments_pb2.DebugInfo:
    return fragments_pb2.DebugInfo(
        ids=[convert_relation_id(key) for key in info.id_to_orig_name.keys()],
        orig_names=info.id_to_orig_name.values()
    )

def convert_define(d: ir.Define) -> transactions_pb2.Define:
    return transactions_pb2.Define(fragment=convert_fragment(d.fragment))

def convert_undefine(u: ir.Undefine) -> transactions_pb2.Undefine:
    return transactions_pb2.Undefine(fragment_id=convert_fragment_id(u.fragment_id))

def convert_context(c: ir.Context) -> transactions_pb2.Context:
    return transactions_pb2.Context(relations=[convert_relation_id(rid) for rid in c.relations])

def convert_write(w: ir.Write) -> transactions_pb2.Write:
    wt = w.write_type
    if isinstance(wt, ir.Define):
        return transactions_pb2.Write(define=convert_define(wt))
    elif isinstance(wt, ir.Undefine):
        return transactions_pb2.Write(undefine=convert_undefine(wt))
    elif isinstance(wt, ir.Context):
        return transactions_pb2.Write(context=convert_context(wt))
    else:
        raise TypeError(f"Unsupported Write type: {type(wt)}")

def convert_demand(d: ir.Demand) -> transactions_pb2.Demand:
    return transactions_pb2.Demand(relation_id=convert_relation_id(d.relation_id))

def convert_output(o: ir.Output) -> transactions_pb2.Output:
    kwargs: Dict[str, Any] = {'relation_id': convert_relation_id(o.relation_id)}
    if o.name is not None:
        kwargs['name'] = o.name
    return transactions_pb2.Output(**kwargs) # type: ignore

def convert_export(e: ir.Export) -> transactions_pb2.Export:
    return transactions_pb2.Export(csv_config=convert_export_config(e.config)) # type: ignore

def convert_export_config(ec: ir.ExportCSVConfig) -> transactions_pb2.ExportCSVConfig:
    return transactions_pb2.ExportCSVConfig(
        data_columns=[convert_export_csv_column(c) for c in ec.data_columns],
        path=ec.path,
        partition_size=ec.partition_size if ec.partition_size is not None else 0,
        compression=ec.compression if ec.compression is not None else "",
        syntax_header_row=ec.syntax_header_row if ec.syntax_header_row is not None else True, # type: ignore
        syntax_missing_string=ec.syntax_missing_string if ec.syntax_missing_string is not None else "",
        syntax_delim=ec.syntax_delim if ec.syntax_delim is not None else ",",
        syntax_quotechar=ec.syntax_quotechar if ec.syntax_quotechar is not None else '"',
        syntax_escapechar=ec.syntax_escapechar if ec.syntax_escapechar is not None else '\\'
    )

def convert_export_csv_column(ec: ir.ExportCSVColumn) -> transactions_pb2.ExportCSVColumn:
    return transactions_pb2.ExportCSVColumn(
        column_name=ec.column_name,
        column_data=convert_relation_id(ec.column_data),
    )

def convert_abort(a: ir.Abort) -> transactions_pb2.Abort:
    kwargs: Dict[str, Any] = {'relation_id': convert_relation_id(a.relation_id)}
    if a.name is not None:
        kwargs['name'] = a.name
    return transactions_pb2.Abort(**kwargs) # type: ignore

def convert_whatif(wi: ir.WhatIf) -> transactions_pb2.WhatIf:
    kwargs: Dict[str, Any] = {'epoch': convert_epoch(wi.epoch)}
    if wi.branch is not None:
        kwargs['branch'] = wi.branch
    return transactions_pb2.WhatIf(**kwargs) # type: ignore

def convert_read(r: ir.Read) -> transactions_pb2.Read:
    rt = r.read_type
    if isinstance(rt, ir.Demand):
        return transactions_pb2.Read(demand=convert_demand(rt))
    elif isinstance(rt, ir.Output):
        return transactions_pb2.Read(output=convert_output(rt))
    elif isinstance(rt, ir.WhatIf):
        return transactions_pb2.Read(what_if=convert_whatif(rt)) # Note the underscore
    elif isinstance(rt, ir.Abort):
        return transactions_pb2.Read(abort=convert_abort(rt))
    elif isinstance(rt, ir.Export):
        return transactions_pb2.Read(export=convert_export(rt))
    else:
        raise TypeError(f"Unsupported Read type: {type(rt)}")

def convert_epoch(e: ir.Epoch) -> transactions_pb2.Epoch:
    return transactions_pb2.Epoch(
        writes=[convert_write(w) for w in e.writes],
        reads=[convert_read(r) for r in e.reads]
    )

def convert_configure(c: ir.Configure) -> transactions_pb2.Configure:
    return transactions_pb2.Configure(
        semantics_version=c.semantics_version,
        ivm_config=convert_ivm_config(c.ivm_config)
    )

def convert_ivm_config(c: ir.IVMConfig) -> transactions_pb2.IVMConfig:
    return transactions_pb2.IVMConfig(
        level=convert_maintenance_level(c.level)
    )

def convert_maintenance_level(l: ir.MaintenanceLevel) -> transactions_pb2.MaintenanceLevel:
    return transactions_pb2.MaintenanceLevel.Name(l.value) # type: ignore[missing-attribute]

def convert_transaction(t: ir.Transaction) -> transactions_pb2.Transaction:
    return transactions_pb2.Transaction(
        configure=convert_configure(t.configure),
        epochs=[convert_epoch(e) for e in t.epochs]
    )

def ir_to_proto(node: ir.LqpNode) -> Union[
    transactions_pb2.Transaction,
    fragments_pb2.Fragment,
    logic_pb2.Declaration,
    logic_pb2.Formula
]:
    if isinstance(node, ir.Transaction):
        return convert_transaction(node)
    elif isinstance(node, ir.Fragment):
        return convert_fragment(node)
    elif isinstance(node, ir.Declaration):
        return convert_declaration(node)
    elif isinstance(node, ir.Formula):
        return convert_formula(node)
    else:
        raise TypeError(f"Unsupported top-level IR node type for conversion: {type(node)}")



import argparse
import os
import hashlib
from dataclasses import dataclass
from lark import Lark, Transformer, v_args
import lqp.ir as ir
from lqp.emit import ir_to_proto
from lqp.validator import validate_lqp
from google.protobuf.json_format import MessageToJson
from decimal import Decimal
from datetime import date, datetime

grammar = """
start: transaction | fragment

transaction: "(transaction" configure? epoch* ")"
configure: "(configure" config_dict ")"
epoch: "(epoch" writes? reads? ")"
writes: "(writes" write* ")"
reads: "(reads" read* ")"

write: define | undefine | context
define: "(define" fragment ")"
undefine: "(undefine" fragment_id ")"
context: "(context" relation_id* ")"

read: demand | output | export | abort
demand: "(demand" relation_id ")"
output: "(output" name? relation_id ")"
export: "(export" export_csv_config ")"
abort: "(abort" name? relation_id ")"

export_csv_config: "(export_csv_config" export_path export_columns config_dict ")"

export_columns: "(columns" export_column* ")"
export_column: "(column" STRING relation_id ")"
export_path: "(path" STRING ")"

fragment: "(fragment" fragment_id declaration* ")"

declaration: def_ | algorithm
def_: "(def" relation_id abstraction attrs? ")"

algorithm: "(algorithm" relation_id* script ")"
script: "(script" construct* ")"

construct: loop | instruction
loop: "(loop" init script ")"
init: "(init" instruction* ")"

instruction: assign | upsert | break_ | copy | monoid_def | monus_def

assign : "(assign" relation_id abstraction attrs? ")"
upsert : "(upsert" relation_id abstraction attrs? ")"
break_ : "(break" relation_id abstraction attrs? ")"
copy : "(copy" relation_id abstraction attrs? ")"
monoid_def : "(monoid" monoid relation_id abstraction attrs? ")"
monus_def : "(monus" monoid relation_id abstraction attrs? ")"

monoid : or_monoid | min_monoid | max_monoid | sum_monoid
or_monoid : "BOOL" "::" "OR"
min_monoid : type_ "::" "MIN"
max_monoid : type_ "::" "MAX"
sum_monoid : type_ "::" "SUM"

abstraction: "(" bindings formula ")"
bindings: "[" binding* "]"
binding: SYMBOL "::" type_

formula: exists | reduce | conjunction | disjunction | not_ | ffi | atom | pragma | primitive | true | false | relatom | cast
exists: "(exists" bindings formula ")"
reduce: "(reduce" abstraction abstraction terms ")"
conjunction: "(and" formula* ")"
disjunction: "(or" formula* ")"
not_: "(not" formula ")"
ffi: "(ffi" name args terms ")"
atom: "(atom" relation_id term* ")"
relatom: "(relatom" name relterm* ")"
cast: "(cast" term term ")"
pragma: "(pragma" name  term* ")"
true: "(true)"
false: "(false)"

args: "(args" abstraction* ")"
terms: "(terms" term* ")"

primitive: raw_primitive | eq | lt | lt_eq | gt | gt_eq | add | minus | multiply | divide
raw_primitive: "(primitive" name relterm* ")"
eq: "(=" term term ")"
lt: "(<" term term ")"
lt_eq: "(<=" term term ")"
gt: "(>" term term ")"
gt_eq: "(>=" term term ")"

add: "(+" term term term ")"
minus: "(-" term term term ")"
multiply: "(*" term term term ")"
divide: "(/" term term term ")"

relterm: specialized_value | term
term: var | constant
specialized_value: "#" value
var: SYMBOL
constant: value

attrs: "(attrs" attribute* ")"
attribute: "(attribute" name constant* ")"

fragment_id: ":" SYMBOL
relation_id: (":" SYMBOL) | NUMBER
name: ":" SYMBOL

value: STRING | NUMBER | FLOAT | UINT128 | INT128
     | date | datetime | MISSING | DECIMAL | BOOLEAN

type_ : TYPE_NAME | "(" TYPE_NAME value* ")"

TYPE_NAME: "STRING" | "INT" | "FLOAT" | "UINT128" | "INT128"
         | "DATE" | "DATETIME" | "MISSING" | "DECIMAL" | "BOOLEAN"

SYMBOL: /[a-zA-Z_][a-zA-Z0-9_.-]*/
MISSING.1: "missing" // Set a higher priority so so it's MISSING instead of SYMBOL
STRING: ESCAPED_STRING
NUMBER: /[-]?\\d+/
INT128: /[-]?\\d+i128/
UINT128: /0x[0-9a-fA-F]+/
FLOAT.1: /[-]?\\d+\\.\\d+/ | "inf" | "nan"
DECIMAL.2: /[-]?\\d+\\.\\d+d\\d+/
BOOLEAN.1: "true" | "false" // Set a higher priority so it's BOOLEAN instead of SYMBOL
date: "(date" NUMBER NUMBER NUMBER ")"
datetime: "(datetime" NUMBER NUMBER NUMBER NUMBER NUMBER NUMBER NUMBER? ")"

config_dict: "{" config_key_value* "}"
config_key_value: ":" SYMBOL value

COMMENT: /;;.*/  // Matches ;; followed by any characters except newline
%ignore /\\s+/
%ignore COMMENT
%import common.ESCAPED_STRING -> ESCAPED_STRING
"""

def construct_configure(config_dict, meta):
    # Construct IVMConfig
    maintenance_level_value = config_dict.get("ivm.maintenance_level")
    if maintenance_level_value:
        maintenance_level = getattr(ir.MaintenanceLevel, maintenance_level_value.value.upper())
    else:
        maintenance_level = ir.MaintenanceLevel.OFF
    ivm_config = ir.IVMConfig(level=maintenance_level, meta=meta)

    # Construct Configure
    semantics_version_value = config_dict.get("semantics_version")
    if semantics_version_value:
        semantics_version = semantics_version_value.value
    else:
        semantics_version = 0

    return ir.Configure(
        semantics_version=semantics_version,
        ivm_config=ivm_config,
        meta=meta,
    )

def desugar_to_raw_primitive(name, terms):
    # Convert terms to relterms
    return ir.Primitive(name=name, terms=terms, meta=None)
@v_args(meta=True)
class LQPTransformer(Transformer):
    def __init__(self, file: str):
        self.file = file
        self.id_to_debuginfo = {}
        self._current_fragment_id = None

    def meta(self, meta):
        return ir.SourceInfo(file=self.file, line=meta.line, column=meta.column)

    def start(self, meta, items):
        return items[0]

    def TYPE_NAME(self, s):
        return getattr(ir.TypeName, s.upper())
    def type_(self, meta, items):
        return ir.Type(type_name=items[0], parameters=items[1:],  meta=self.meta(meta))


    #
    # Transactions
    #
    def transaction(self, meta, items):
        if isinstance(items[0], ir.Configure):
            configure = items[0]
            epochs = items[1:]
        else:
            configure = construct_configure({}, self.meta(meta))
            epochs = items
            
        return ir.Transaction(configure=configure, epochs=epochs, meta=self.meta(meta))

    def configure(self, meta, items):
        return construct_configure(items[0], self.meta(meta))

    def epoch(self, meta, items):
        kwargs = {k: v for k, v in items if v} # Filter out None values
        return ir.Epoch(**kwargs, meta=self.meta(meta))

    def writes(self, meta, items):
        return ("writes", items)
    def reads(self, meta, items):
        return ("reads", items)
    def write(self, meta, items):
        return ir.Write(write_type=items[0], meta=self.meta(meta))

    def define(self, meta, items):
        return ir.Define(fragment=items[0], meta=self.meta(meta))

    def undefine(self, meta, items):
        return ir.Undefine(fragment_id=items[0], meta=self.meta(meta))

    def context(self, meta, items):
        return ir.Context(relations=items, meta=self.meta(meta))

    def read(self, meta, items):
        return ir.Read(read_type=items[0], meta=self.meta(meta))
    def demand(self, meta, items):
        return ir.Demand(relation_id=items[0], meta=self.meta(meta))

    def output(self, meta, items):
        if len(items) == 1:
            return ir.Output(name=None, relation_id=items[0], meta=self.meta(meta))
        return ir.Output(name=items[0], relation_id=items[1], meta=self.meta(meta))

    def export(self, meta, items):
        return ir.Export(config=items[0], meta=self.meta(meta))

    def export_csv_config(self, meta, items):
        assert len(items) >= 2, "Export config must have at least columns and path"

        export_fields = {}
        for i in items[2:]:
            assert isinstance(i, dict)
            for k, v in i.items():
                export_fields[k] = v.value

        return ir.ExportCSVConfig(
            path=items[0],
            data_columns=items[1],
            **export_fields,
            meta=self.meta(meta)
        )

    def export_columns(self, meta, items):
        # items is a list of ExportCSVColumn objects
        return items

    def export_column(self, meta, items):
        return ir.ExportCSVColumn(
            column_name=items[0],
            column_data=items[1],
            meta=self.meta(meta)
        )

    def export_path(self, meta, items):
        return items[0]

    def abort(self, meta, items):
        if len(items) == 1:
            return ir.Abort(name=None, relation_id=items[0], meta=self.meta(meta))
        return ir.Abort(name=items[0], relation_id=items[1], meta=self.meta(meta))

    #
    # Logic
    #
    def fragment(self, meta, items):
        fragment_id = items[0]
        debug_info = ir.DebugInfo(id_to_orig_name=self.id_to_debuginfo[fragment_id], meta=self.meta(meta))
        self._current_fragment_id = None
        return ir.Fragment(id=fragment_id, declarations=items[1:], debug_info=debug_info, meta=self.meta(meta))

    def fragment_id(self, meta, items):
        fragment_id = ir.FragmentId(id=items[0].encode(), meta=self.meta(meta))
        self._current_fragment_id = fragment_id # type: ignore
        if fragment_id not in self.id_to_debuginfo:
            self.id_to_debuginfo[fragment_id] = {}
        return fragment_id

    def declaration(self, meta, items):
        return items[0]
    def def_(self, meta, items):
        name = items[0]
        body = items[1]
        attrs = items[2] if len(items) > 2 else []
        return ir.Def(name=name, body=body, attrs=attrs, meta=self.meta(meta))

    def algorithm(self, meta, items):
        return ir.Algorithm(global_=items[:-1], body=items[-1], meta=self.meta(meta))
    def script(self, meta, items):
        return ir.Script(constructs=items, meta=self.meta(meta))

    def construct(self, meta, items):
        return items[0]

    def loop(self, meta, items):
        init = items[0]
        script = items[1]
        return ir.Loop(init=init, body=script, meta=self.meta(meta))
    def init(self, meta, items):
        return items

    def instruction(self, meta, items):
        return items[0]

    def assign(self, meta, items):
        name = items[0]
        body = items[1]
        attrs = items[2] if len(items) > 2 else []
        return ir.Assign(name=name, body=body, attrs=attrs, meta=self.meta(meta))
    def upsert(self, meta, items):
        name = items[0]
        body = items[1]
        attrs = items[2] if len(items) > 2 else []
        return ir.Upsert(name=name, body=body, attrs=attrs, meta=self.meta(meta))
    def break_(self, meta, items):
        name = items[0]
        body = items[1]
        attrs = items[2] if len(items) > 2 else []
        return ir.Break(name=name, body=body, attrs=attrs, meta=self.meta(meta))
    def copy(self, meta, items):
        name = items[0]
        body = items[1]
        attrs = items[2] if len(items) > 2 else []
        return ir.Copy(name=name, body=body, attrs=attrs, meta=self.meta(meta))
    def monoid_def(self, meta, items):
        monoid = items[0]
        name = items[1]
        body = items[2]
        attrs = items[3] if len(items) > 3 else []
        return ir.MonoidDef(monoid=monoid, name=name, body=body, attrs=attrs, meta=self.meta(meta))
    def monus_def(self, meta, items):
        monoid = items[0]
        name = items[1]
        body = items[2]
        attrs = items[3] if len(items) > 3 else []
        return ir.MonusDef(monoid=monoid, name=name, body=body, attrs=attrs, meta=self.meta(meta))

    def monoid(self, meta, items) :
        return items[0]
    def or_monoid(self, meta, items):
        return ir.OrMonoid(meta=meta)
    def min_monoid(self, meta, items):
        return ir.MinMonoid(type=items[0], meta=meta)
    def max_monoid(self, meta, items):
        return ir.MaxMonoid(type=items[0], meta=meta)
    def sum_monoid(self, meta, items):
        return ir.SumMonoid(type=items[0], meta=meta)

    def abstraction(self, meta, items):
        return ir.Abstraction(vars=items[0], value=items[1], meta=self.meta(meta))

    def binding(self, meta, items):
        name, rel_t = items
        return (ir.Var(name=name, meta=self.meta(meta)), rel_t)

    def vars(self, meta, items):
        return items
    def bindings(self, meta, items):
        return items
    def attrs(self, meta, items):
        return items

    def formula(self, meta, items):
        return items[0]
    def true(self, _, meta):
        return ir.Conjunction(args=[], meta=self.meta(meta))

    def false(self, _, meta):
        return ir.Disjunction(args=[], meta=self.meta(meta))

    def exists(self, meta, items):
        # Create Abstraction for body directly here
        body_abstraction = ir.Abstraction(vars=items[0], value=items[1], meta=self.meta(meta))
        return ir.Exists(body=body_abstraction, meta=self.meta(meta))

    def reduce(self, meta, items):
        return ir.Reduce(op=items[0], body=items[1], terms=items[2], meta=self.meta(meta))

    def conjunction(self, meta, items):
        return ir.Conjunction(args=items, meta=self.meta(meta))

    def disjunction(self, meta, items):
        return ir.Disjunction(args=items, meta=self.meta(meta))

    def not_(self, meta, items):
        return ir.Not(arg=items[0], meta=self.meta(meta))

    def ffi(self, meta, items):
        return ir.FFI(name=items[0], args=items[1], terms=items[2], meta=self.meta(meta))

    def atom(self, meta, items):
        return ir.Atom(name=items[0], terms=items[1:], meta=self.meta(meta))

    def pragma(self, meta, items):
        return ir.Pragma(name=items[0], terms=items[1:], meta=self.meta(meta))

    def relatom(self, meta, items):
        return ir.RelAtom(name=items[0], terms=items[1:], meta=self.meta(meta))

    def cast(self, meta, items):
        return ir.Cast(input=items[0], result=items[1], meta=self.meta(meta))

    #
    # Primitives
    #
    def primitive(self, meta, items):
        if isinstance(items[0], ir.Formula):
            return items[0]
        raise TypeError(f"Unexpected primitive type: {type(items[0])}")
    def raw_primitive(self, meta, items):
        return ir.Primitive(name=items[0], terms=items[1:], meta=self.meta(meta))
    def _make_primitive(self, name_symbol, terms, meta):
         # Convert name symbol to string if needed, assuming self.name handles it
         name_str = self.name([name_symbol], meta) if isinstance(name_symbol, str) else name_symbol
         return ir.Primitive(name=name_str, terms=terms, meta=self.meta(meta))
    def eq(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_eq"]), items)
    def lt(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_lt_monotype"]), items)
    def lt_eq(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_lt_eq_monotype"]), items)
    def gt(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_gt_monotype"]), items)
    def gt_eq(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_gt_eq_monotype"]), items)

    def add(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_add_monotype"]), items)
    def minus(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_subtract_monotype"]), items)
    def multiply(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_multiply_monotype"]), items)
    def divide(self, meta, items):
        return desugar_to_raw_primitive(self.name(meta, ["rel_primitive_divide_monotype"]), items)

    def args(self, meta, items):
        return items
    def terms(self, meta, items):
        return items

    def relterm(self, meta, items):
        return items[0]
    def term(self, meta, items):
        return items[0]
    def var(self, meta, items):
        return ir.Var(name=items[0], meta=self.meta(meta))
    def constant(self, meta, items):
        return items[0]
    def specialized_value(self, meta, items):
        return ir.SpecializedValue(value=items[0], meta=self.meta(meta))

    def name(self, meta, items):
        return items[0]

    def attribute(self, meta, items):
        return ir.Attribute(name=items[0], args=items[1:], meta=self.meta(meta))

    def relation_id(self, meta, items):
        ident = items[0] # Remove leading ':'
        if isinstance(ident, str):
            # First 64 bits of SHA-256 as the id
            id_val = int(hashlib.sha256(ident.encode()).hexdigest()[:16], 16)
            result = ir.RelationId(id=id_val, meta=self.meta(meta))

            # Store mapping in the current fragment's debug info
            if self._current_fragment_id is not None:
                self.id_to_debuginfo[self._current_fragment_id][result] = ident
            return result

        elif isinstance(ident, int):
            return ir.RelationId(id=ident, meta=self.meta(meta))

    #
    # Primitive values
    #
    def value(self, meta, items):
        return ir.Value(value=items[0], meta=self.meta(meta))

    def STRING(self, s):
        return s[1:-1].encode().decode('unicode_escape') # Strip quotes and process escaping
    def NUMBER(self, n):
        return int(n)
    def FLOAT(self, f):
        return float(f)
    def SYMBOL(self, sym):
        return str(sym)
    def UINT128(self, u):
        uint128_val = int(u, 16)
        return ir.UInt128Value(value=uint128_val, meta=None)
    def INT128(self, u):
        u= u[:-4]  # Remove the 'i128' suffix
        int128_val = int(u)
        return ir.Int128Value(value=int128_val, meta=None)
    def MISSING(self, m):
        return ir.MissingValue(meta=None)
    def DECIMAL(self, d):
        # Decimal is a string like "123.456d12" where the last part after `d` is the
        # precision, and the scale is the number of digits between the decimal point and `d`
        parts = d.split('d')
        if len(parts) != 2:
            raise ValueError(f"Invalid decimal format: {d}")
        scale = len(parts[0].split('.')[1])
        precision = int(parts[1])
        value = Decimal(parts[0])

        return ir.DecimalValue(precision=precision, scale=scale, value=value, meta=None)
    def BOOLEAN(self, b):
        return ir.BooleanValue(value=bool(b == "true"), meta=None)
    def date(self, meta, items):
        # Date is in the format (date YYYY MM DD)
        date_val = date(*items)
        return ir.DateValue(value=date_val, meta=None)
    def datetime(self, meta, items):
        # Date is in the format (datetime YYYY MM DD HH MM SS [MS])
        datetime_val = datetime(*items)
        return ir.DateTimeValue(value=datetime_val, meta=None)

    def config_dict(self, meta, items):
        # items is a list of key-value pairs
        config = {}
        for (k, v) in items:
            config[k] = v
        return config

    def config_key_value(self, meta, items):
        assert len(items) == 2
        return (items[0], items[1])

# LALR(1) is significantly faster than Earley for parsing, especially on larger inputs. It
# uses a precomputed parse table, reducing runtime complexity to O(n) (linear in input
# size), whereas Earley is O(n³) in the worst case (though often O(n²) or better for
# practical grammars). The LQP grammar is relatively complex but unambiguous, making
# LALR(1)’s speed advantage appealing for a CLI tool where quick parsing matters.
parser = Lark(grammar, parser="lalr", propagate_positions=True)

def parse_lqp(file, text) -> ir.LqpNode:
    """Parse LQP text and return an IR node that can be converted to protocol buffers"""
    tree = parser.parse(text)
    transformer = LQPTransformer(file)
    result = transformer.transform(tree)
    return result

def process_file(filename, bin, json):
    with open(filename, "r") as f:
        lqp_text = f.read()

    lqp = parse_lqp(filename, lqp_text)
    validate_lqp(lqp) # type: ignore
    lqp_proto = ir_to_proto(lqp)
    print(lqp_proto)

    # Write binary output to the configured directories, using the same filename.
    if bin:
        with open(bin, "wb") as f:
            f.write(lqp_proto.SerializeToString())

    # Write JSON output
    if json:
        with open(json, "w") as f:
            f.write(MessageToJson(lqp_proto, preserving_proto_field_name=True))

def main():
    arg_parser = argparse.ArgumentParser(description="Parse LQP S-expression into Protobuf binary and JSON files.")
    arg_parser.add_argument("input_directory", help="path to the input LQP S-expression files")
    arg_parser.add_argument("--bin", help="output directory for the binary encoded protobuf")
    arg_parser.add_argument("--json", help="output directory for the JSON encoded protobuf")
    args = arg_parser.parse_args()

    print(args)

    # Check if directory
    if not os.path.isdir(args.input_directory):
        filename = args.input_directory
        if not filename.endswith(".lqp"):
            print(f"Skipping file {filename} as it does not have the .lqp extension")
            return

        bin = args.bin if args.bin else None
        if bin and not bin.endswith(".bin"):
            print(f"Skipping output {bin} as it does not have the .bin extension")
            bin = None

        json = args.json if args.json else None
        if json and not json.endswith(".json"):
            print(f"Skipping output {json} as it does not have the .json extension")
            json = None
        process_file(filename, bin, json)

    else:
        # Process each file in the input directory
        for file in os.listdir(args.input_directory):
            if not file.endswith(".lqp"):
                print(f"Skipping file {file} as it does not have the .lqp extension")
                continue

            filename = os.path.join(args.input_directory, file)
            basename = os.path.splitext(file)[0]
            bin = os.path.join(args.bin, basename+".bin") if args.bin else None
            json = os.path.join(args.json, basename+".json") if args.json else None
            process_file(filename, bin, json)


if __name__ == "__main__":
    main()
